// This is a re-implementation of new, using just the reflection tools available in JS.
//
// Usage: instead of
//   new SomeConstructor(foo, bar)
// call
//   fauxNew(SomeConstructor, [foo, bar])
//
function fauxNew (constructor, args) {
    // allocate a new object and set the prototype
    var newObject = Object.create(constructor.prototype)
    // call the constructoir with "this" bound to the new object
    var retVal = constructor.apply(newObject, args || [])
    // if the constructor returned an object, return it; 
    // otherwise return the new object
    var constructorReturnedAnObject = 
        !!retVal && ["object", "function"].indexOf(typeof retVal) !== -1
    return constructorReturnedAnObject? retVal : newObject
};

function RegularClass (foo) { 
    this.foo = foo
}

function ReturnsAnObject (foo) {
    this.foo = foo
    return {bar: "i am not this"}
}

function ReturnsANumber (foo) {
    this.foo = foo
    return 5
}

function ReturnsAString (foo) {
    this.foo = foo
    return "five"
}

// dummy function for use in ReturnsAFunction
function noop () {}

function ReturnsAFunction (foo) {
    this.foo = foo
    return noop
}

function HasAPrototype (foo) {
    this.foo = foo
}

HasAPrototype.prototype = {
    bar: "i come from the prototype"
}

function ChecksInstanceOf () {
    this.isInstance = this instanceof ChecksInstanceOf
}
;

function assert (predicate, message) {
    console.log((predicate? "SUCCESS: ":"FAIL: ") + message)
}

// doesn't matter what this is, we just need it to be a consistent 
// value that can be passed in and out
var testArg = Symbol("I am a test argument")

function test (constructor, mapper, expected, message) {
    var traditional = mapper(new constructor(testArg))
    var faux = mapper(fauxNew(constructor, [testArg]))
    assert (faux == traditional && faux == expected, message)
};

test(RegularClass,    x => x.foo, testArg,         "member is set correctly")
test(ReturnsAnObject, x => x.bar, "i am not this", "object is returned correctly")
test(ReturnsANumber,  x => x.foo, testArg,         "returned number is ignored")
;
test(ReturnsAString,   x => x.foo, testArg, "returned string is ignored")
test(ReturnsAFunction, x => x,     noop,    "function is returned correctly")
test(HasAPrototype,    x => x.bar, "i come from the prototype", "prototype is applied correctly");
test(ChecksInstanceOf, x => x.isInstance, true, "passes instanceof test");
